@@ -30,6 +30,16 @@
 #include "gstinfo.h"
 #include "gsterror.h"
 
+static GstStaticPadTemplate sinktemplate = GST_STATIC_PAD_TEMPLATE ("sink",
+    GST_PAD_SINK,
+    GST_PAD_ALWAYS,
+    GST_STATIC_CAPS_ANY);
+
+static GstStaticPadTemplate srctemplate = GST_STATIC_PAD_TEMPLATE ("src",
+    GST_PAD_SRC,
+    GST_PAD_ALWAYS,
+    GST_STATIC_CAPS_ANY);
+
 GST_DEBUG_CATEGORY_STATIC (queue_dataflow);
 
 static GstElementDetails gst_queue_details = GST_ELEMENT_DETAILS ("Queue",
@@ -67,6 +77,24 @@ enum
       /* FILL ME */
 };
 
+#define GST_QUEUE_MUTEX_LOCK G_STMT_START {				\
+  GST_CAT_LOG_OBJECT (queue_dataflow, queue,				\
+      "locking qlock from thread %p",					\
+      g_thread_self ());						\
+  g_mutex_lock (queue->qlock);						\
+  GST_CAT_LOG_OBJECT (queue_dataflow, queue,				\
+      "locked qlock from thread %p",					\
+      g_thread_self ());						\
+} G_STMT_END
+
+#define GST_QUEUE_MUTEX_UNLOCK G_STMT_START {				\
+  GST_CAT_LOG_OBJECT (queue_dataflow, queue,				\
+      "unlocking qlock from thread %p",					\
+      g_thread_self ());						\
+  g_mutex_unlock (queue->qlock);					\
+} G_STMT_END
+
+
 typedef struct _GstQueueEventResponse
 {
   GstEvent *event;
@@ -77,7 +105,7 @@ GstQueueEventResponse;
 static void gst_queue_base_init (GstQueueClass * klass);
 static void gst_queue_class_init (GstQueueClass * klass);
 static void gst_queue_init (GstQueue * queue);
-static void gst_queue_dispose (GObject * object);
+static void gst_queue_finalize (GObject * object);
 
 static void gst_queue_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec);
@@ -152,6 +180,10 @@ gst_queue_base_init (GstQueueClass * klass)
 {
   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (klass);
 
+  gst_element_class_add_pad_template (gstelement_class,
+      gst_static_pad_template_get (&srctemplate));
+  gst_element_class_add_pad_template (gstelement_class,
+      gst_static_pad_template_get (&sinktemplate));
   gst_element_class_set_details (gstelement_class, &gst_queue_details);
 }
 
@@ -232,7 +264,7 @@ gst_queue_class_init (GstQueueClass * klass)
           0, G_MAXUINT64, -1, G_PARAM_READWRITE));
 
   /* set several parent class virtual functions */
-  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_queue_dispose);
+  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_queue_finalize);
   gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_queue_set_property);
   gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_queue_get_property);
 
@@ -247,7 +279,9 @@ gst_queue_init (GstQueue * queue)
   GST_FLAG_SET (queue, GST_ELEMENT_DECOUPLED);
   GST_FLAG_SET (queue, GST_ELEMENT_EVENT_AWARE);
 
-  queue->sinkpad = gst_pad_new ("sink", GST_PAD_SINK);
+  queue->sinkpad =
+      gst_pad_new_from_template (gst_static_pad_template_get (&sinktemplate),
+      "sink");
   gst_pad_set_chain_function (queue->sinkpad,
       GST_DEBUG_FUNCPTR (gst_queue_chain));
   gst_element_add_pad (GST_ELEMENT (queue), queue->sinkpad);
@@ -257,7 +291,9 @@ gst_queue_init (GstQueue * queue)
       GST_DEBUG_FUNCPTR (gst_queue_getcaps));
   gst_pad_set_active (queue->sinkpad, TRUE);
 
-  queue->srcpad = gst_pad_new ("src", GST_PAD_SRC);
+  queue->srcpad =
+      gst_pad_new_from_template (gst_static_pad_template_get (&srctemplate),
+      "src");
   gst_pad_set_get_function (queue->srcpad, GST_DEBUG_FUNCPTR (gst_queue_get));
   gst_element_add_pad (GST_ELEMENT (queue), queue->srcpad);
   gst_pad_set_link_function (queue->srcpad, GST_DEBUG_FUNCPTR (gst_queue_link));
@@ -288,18 +324,20 @@ gst_queue_init (GstQueue * queue)
   queue->item_del = g_cond_new ();
   queue->event_done = g_cond_new ();
   queue->events = g_queue_new ();
+  queue->event_lock = g_mutex_new ();
   queue->queue = g_queue_new ();
 
   GST_CAT_DEBUG_OBJECT (GST_CAT_THREAD, queue,
       "initialized queue's not_empty & not_full conditions");
 }
 
+/* called only once, as opposed to dispose */
 static void
-gst_queue_dispose (GObject * object)
+gst_queue_finalize (GObject * object)
 {
   GstQueue *queue = GST_QUEUE (object);
 
-  gst_element_set_state (GST_ELEMENT (queue), GST_STATE_NULL);
+  GST_DEBUG_OBJECT (queue, "finalizing queue");
 
   while (!g_queue_is_empty (queue->queue)) {
     GstData *data = g_queue_pop_head (queue->queue);
@@ -311,14 +349,18 @@ gst_queue_dispose (GObject * object)
   g_cond_free (queue->item_add);
   g_cond_free (queue->item_del);
   g_cond_free (queue->event_done);
+  g_mutex_lock (queue->event_lock);
   while (!g_queue_is_empty (queue->events)) {
-    GstEvent *event = g_queue_pop_head (queue->events);
+    GstQueueEventResponse *er = g_queue_pop_head (queue->events);
 
-    gst_event_unref (event);
+    gst_event_unref (er->event);
   }
+  g_mutex_unlock (queue->event_lock);
+  g_mutex_free (queue->event_lock);
+  g_queue_free (queue->events);
 
-  if (G_OBJECT_CLASS (parent_class)->dispose)
-    G_OBJECT_CLASS (parent_class)->dispose (object);
+  if (G_OBJECT_CLASS (parent_class)->finalize)
+    G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static GstCaps *
@@ -389,15 +431,34 @@ static void
 gst_queue_handle_pending_events (GstQueue * queue)
 {
   /* check for events to send upstream */
+  /* g_queue_get_length is glib 2.4, so don't depend on it yet, use ->length */
+  GST_CAT_DEBUG_OBJECT (queue_dataflow, queue,
+      "handling pending events, events queue of size %d",
+      queue->events->length);
+  g_mutex_lock (queue->event_lock);
   while (!g_queue_is_empty (queue->events)) {
-    GstQueueEventResponse *er = g_queue_pop_head (queue->events);
+    GstQueueEventResponse *er;
 
-    GST_CAT_DEBUG_OBJECT (queue_dataflow, queue, "sending event upstream");
+    er = g_queue_pop_head (queue->events);
+
+    GST_CAT_DEBUG_OBJECT (queue_dataflow, queue,
+        "sending event %p (%d) from event response %p upstream",
+        er->event, GST_EVENT_TYPE (er->event), er);
+    if (er->handled) {
+      /* change this to an assert when this file gets reviewed properly. */
+      GST_ELEMENT_ERROR (queue, CORE, EVENT, (NULL),
+          ("already handled event %p (%d) from event response %p upstream",
+              er->event, GST_EVENT_TYPE (er->event), er));
+      break;
+    }
+    g_mutex_unlock (queue->event_lock);
     er->ret = gst_pad_event_default (queue->srcpad, er->event);
     er->handled = TRUE;
     g_cond_signal (queue->event_done);
+    g_mutex_lock (queue->event_lock);
     GST_CAT_DEBUG_OBJECT (queue_dataflow, queue, "event sent");
   }
+  g_mutex_unlock (queue->event_lock);
 }
 
 #define STATUS(queue, msg) \
@@ -430,9 +491,7 @@ gst_queue_chain (GstPad * pad, GstData * data)
 
 restart:
   /* we have to lock the queue since we span threads */
-  GST_CAT_LOG_OBJECT (queue_dataflow, queue, "locking t:%p", g_thread_self ());
-  g_mutex_lock (queue->qlock);
-  GST_CAT_LOG_OBJECT (queue_dataflow, queue, "locked t:%p", g_thread_self ());
+  GST_QUEUE_MUTEX_LOCK;
 
   gst_queue_handle_pending_events (queue);
 
@@ -471,9 +530,9 @@ restart:
               queue->cur_level.bytes >= queue->max_size.bytes) ||
           (queue->max_size.time > 0 &&
               queue->cur_level.time >= queue->max_size.time))) {
-    g_mutex_unlock (queue->qlock);
+    GST_QUEUE_MUTEX_UNLOCK;
     g_signal_emit (G_OBJECT (queue), gst_queue_signals[SIGNAL_OVERRUN], 0);
-    g_mutex_lock (queue->qlock);
+    GST_QUEUE_MUTEX_LOCK;
 
     /* how are we going to make space for this buffer? */
     switch (queue->leaky) {
@@ -482,7 +541,7 @@ restart:
         GST_CAT_DEBUG_OBJECT (queue_dataflow, queue,
             "queue is full, leaking buffer on upstream end");
         /* now we can clean up and exit right away */
-        g_mutex_unlock (queue->qlock);
+        GST_QUEUE_MUTEX_UNLOCK;
         goto out_unref;
 
         /* leak first buffer in the queue */
@@ -545,7 +604,7 @@ restart:
            * half of state change executes */
           if (queue->interrupt) {
             GST_CAT_DEBUG_OBJECT (queue_dataflow, queue, "interrupted");
-            g_mutex_unlock (queue->qlock);
+            GST_QUEUE_MUTEX_UNLOCK;
             if (gst_scheduler_interrupt (gst_pad_get_scheduler (queue->sinkpad),
                     GST_ELEMENT (queue))) {
               goto out_unref;
@@ -567,7 +626,7 @@ restart:
             /* this means the other end is shut down. Try to
              * signal to resolve the error */
             if (!queue->may_deadlock) {
-              g_mutex_unlock (queue->qlock);
+              GST_QUEUE_MUTEX_UNLOCK;
               gst_data_unref (data);
               GST_ELEMENT_ERROR (queue, CORE, THREAD, (NULL),
                   ("deadlock found, shutting down source pad elements"));
@@ -589,15 +648,15 @@ restart:
            * that, we handle pending upstream events here, too. */
           gst_queue_handle_pending_events (queue);
 
-          STATUS (queue, "waiting for item_del signal");
+          STATUS (queue, "waiting for item_del signal from thread using qlock");
           g_cond_wait (queue->item_del, queue->qlock);
-          STATUS (queue, "received item_del signal");
+          STATUS (queue, "received item_del signal from thread using qlock");
         }
 
         STATUS (queue, "post-full wait");
-        g_mutex_unlock (queue->qlock);
+        GST_QUEUE_MUTEX_UNLOCK;
         g_signal_emit (G_OBJECT (queue), gst_queue_signals[SIGNAL_RUNNING], 0);
-        g_mutex_lock (queue->qlock);
+        GST_QUEUE_MUTEX_LOCK;
         break;
     }
   }
@@ -621,7 +680,7 @@ restart:
 
   GST_CAT_LOG_OBJECT (queue_dataflow, queue, "signalling item_add");
   g_cond_signal (queue->item_add);
-  g_mutex_unlock (queue->qlock);
+  GST_QUEUE_MUTEX_UNLOCK;
 
   return;
 
@@ -643,9 +702,7 @@ gst_queue_get (GstPad * pad)
 
 restart:
   /* have to lock for thread-safety */
-  GST_CAT_LOG_OBJECT (queue_dataflow, queue, "locking t:%p", g_thread_self ());
-  g_mutex_lock (queue->qlock);
-  GST_CAT_LOG_OBJECT (queue_dataflow, queue, "locked t:%p", g_thread_self ());
+  GST_QUEUE_MUTEX_LOCK;
 
   if (queue->queue->length == 0 ||
       (queue->min_threshold.buffers > 0 &&
@@ -654,9 +711,9 @@ restart:
           queue->cur_level.bytes < queue->min_threshold.bytes) ||
       (queue->min_threshold.time > 0 &&
           queue->cur_level.time < queue->min_threshold.time)) {
-    g_mutex_unlock (queue->qlock);
+    GST_QUEUE_MUTEX_UNLOCK;
     g_signal_emit (G_OBJECT (queue), gst_queue_signals[SIGNAL_UNDERRUN], 0);
-    g_mutex_lock (queue->qlock);
+    GST_QUEUE_MUTEX_LOCK;
 
     STATUS (queue, "pre-empty wait");
     while (queue->queue->length == 0 ||
@@ -671,7 +728,7 @@ restart:
        * change executes. */
       if (queue->interrupt) {
         GST_CAT_DEBUG_OBJECT (queue_dataflow, queue, "interrupted");
-        g_mutex_unlock (queue->qlock);
+        GST_QUEUE_MUTEX_UNLOCK;
         if (gst_scheduler_interrupt (gst_pad_get_scheduler (queue->srcpad),
                 GST_ELEMENT (queue)))
           return GST_DATA (gst_event_new (GST_EVENT_INTERRUPT));
@@ -680,7 +737,7 @@ restart:
       if (GST_STATE (queue) != GST_STATE_PLAYING) {
         /* this means the other end is shut down */
         if (!queue->may_deadlock) {
-          g_mutex_unlock (queue->qlock);
+          GST_QUEUE_MUTEX_UNLOCK;
           GST_ELEMENT_ERROR (queue, CORE, THREAD, (NULL),
               ("deadlock found, shutting down sink pad elements"));
           goto restart;
@@ -698,22 +755,28 @@ restart:
 
         g_get_current_time (&timeout);
         g_time_val_add (&timeout, queue->block_timeout / 1000);
+        GST_LOG_OBJECT (queue, "g_cond_time_wait using qlock from thread %p",
+            g_thread_self ());
         if (!g_cond_timed_wait (queue->item_add, queue->qlock, &timeout)) {
-          g_mutex_unlock (queue->qlock);
+          GST_QUEUE_MUTEX_UNLOCK;
           GST_CAT_WARNING_OBJECT (queue_dataflow, queue,
               "Sending filler event");
           return GST_DATA (gst_event_new_filler ());
         }
       } else {
+        GST_LOG_OBJECT (queue, "doing g_cond_wait using qlock from thread %p",
+            g_thread_self ());
         g_cond_wait (queue->item_add, queue->qlock);
+        GST_LOG_OBJECT (queue, "done g_cond_wait using qlock from thread %p",
+            g_thread_self ());
       }
       STATUS (queue, "got item_add signal");
     }
 
     STATUS (queue, "post-empty wait");
-    g_mutex_unlock (queue->qlock);
+    GST_QUEUE_MUTEX_UNLOCK;
     g_signal_emit (G_OBJECT (queue), gst_queue_signals[SIGNAL_RUNNING], 0);
-    g_mutex_lock (queue->qlock);
+    GST_QUEUE_MUTEX_LOCK;
   }
 
   /* There's something in the list now, whatever it is */
@@ -740,7 +803,7 @@ restart:
 
   GST_CAT_LOG_OBJECT (queue_dataflow, queue, "signalling item_del");
   g_cond_signal (queue->item_del);
-  g_mutex_unlock (queue->qlock);
+  GST_QUEUE_MUTEX_UNLOCK;
 
   /* FIXME: I suppose this needs to be locked, since the EOS
    * bit affects the pipeline state. However, that bit is
@@ -769,7 +832,9 @@ gst_queue_handle_src_event (GstPad * pad, GstEvent * event)
   GstQueue *queue = GST_QUEUE (gst_pad_get_parent (pad));
   gboolean res;
 
-  g_mutex_lock (queue->qlock);
+  GST_CAT_DEBUG_OBJECT (queue_dataflow, queue, "got event %p (%d)",
+      event, GST_EVENT_TYPE (event));
+  GST_QUEUE_MUTEX_LOCK;
 
   if (gst_element_get_state (GST_ELEMENT (queue)) == GST_STATE_PLAYING) {
     GstQueueEventResponse er;
@@ -777,7 +842,12 @@ gst_queue_handle_src_event (GstPad * pad, GstEvent * event)
     /* push the event to the queue and wait for upstream consumption */
     er.event = event;
     er.handled = FALSE;
+    g_mutex_lock (queue->event_lock);
+    GST_CAT_DEBUG_OBJECT (queue_dataflow, queue,
+        "putting event %p (%d) on internal queue", event,
+        GST_EVENT_TYPE (event));
     g_queue_push_tail (queue->events, &er);
+    g_mutex_unlock (queue->event_lock);
     GST_CAT_WARNING_OBJECT (queue_dataflow, queue,
         "Preparing for loop for event handler");
     /* see the chain function on why this is here - it prevents a deadlock */
@@ -787,16 +857,22 @@ gst_queue_handle_src_event (GstPad * pad, GstEvent * event)
 
       g_get_current_time (&timeout);
       g_time_val_add (&timeout, 500 * 1000);    /* half a second */
+      GST_LOG_OBJECT (queue, "doing g_cond_wait using qlock from thread %p",
+          g_thread_self ());
       if (!g_cond_timed_wait (queue->event_done, queue->qlock, &timeout) &&
           !er.handled) {
         GST_CAT_WARNING_OBJECT (queue_dataflow, queue,
-            "timeout in upstream event handling");
-        /* remove ourselves from the pending list. Since we're
-         * locked, others cannot reference this anymore. */
-        queue->queue->head = g_list_remove (queue->queue->head, &er);
-        queue->queue->head = g_list_first (queue->queue->head);
-        queue->queue->tail = g_list_last (queue->queue->head);
-        queue->queue->length--;
+            "timeout in upstream event handling, dropping event %p (%d)",
+            er.event, GST_EVENT_TYPE (er.event));
+        g_mutex_lock (queue->event_lock);
+        /* since this queue is for src events (ie upstream), this thread is
+         * the only one that is pushing stuff on it, so we're sure that
+         * it's still the tail element.  FIXME: But in practice, we should use
+         * GList instead of GQueue for this so we can remove any element in
+         * the list. */
+        g_queue_pop_tail (queue->events);
+        g_mutex_unlock (queue->event_lock);
+        gst_event_unref (er.event);
         res = FALSE;
         goto handled;
       }
@@ -821,7 +897,7 @@ gst_queue_handle_src_event (GstPad * pad, GstEvent * event)
     }
   }
 handled:
-  g_mutex_unlock (queue->qlock);
+  GST_QUEUE_MUTEX_UNLOCK;
 
   return res;
 }
@@ -833,11 +909,11 @@ gst_queue_release_locks (GstElement * element)
 
   queue = GST_QUEUE (element);
 
-  g_mutex_lock (queue->qlock);
+  GST_QUEUE_MUTEX_LOCK;
   queue->interrupt = TRUE;
   g_cond_signal (queue->item_add);
   g_cond_signal (queue->item_del);
-  g_mutex_unlock (queue->qlock);
+  GST_QUEUE_MUTEX_UNLOCK;
 
   return TRUE;
 }
@@ -855,7 +931,7 @@ gst_queue_change_state (GstElement * element)
   /* lock the queue so another thread (not in sync with this thread's state)
    * can't call this queue's _get (or whatever)
    */
-  g_mutex_lock (queue->qlock);
+  GST_QUEUE_MUTEX_LOCK;
 
   switch (GST_STATE_TRANSITION (element)) {
     case GST_STATE_NULL_TO_READY:
@@ -869,7 +945,7 @@ gst_queue_change_state (GstElement * element)
         g_cond_signal (queue->item_add);
 
         ret = GST_STATE_FAILURE;
-        goto error;
+        goto unlock;
       } else {
         GstScheduler *src_sched, *sink_sched;
 
@@ -885,13 +961,14 @@ gst_queue_change_state (GstElement * element)
               GST_ELEMENT_NAME (queue));
 
           ret = GST_STATE_FAILURE;
-          goto error;
+          goto unlock;
         }
       }
       queue->interrupt = FALSE;
       break;
     case GST_STATE_PAUSED_TO_READY:
       gst_queue_locked_flush (queue);
+      gst_caps_replace (&queue->negotiated_caps, NULL);
       break;
     default:
       break;
@@ -906,8 +983,8 @@ gst_queue_change_state (GstElement * element)
   gst_pad_set_active (queue->sinkpad, TRUE);
   gst_pad_set_active (queue->srcpad, TRUE);
 
-error:
-  g_mutex_unlock (queue->qlock);
+unlock:
+  GST_QUEUE_MUTEX_UNLOCK;
 
   GST_CAT_LOG_OBJECT (GST_CAT_STATES, element, "done with state change");
 
@@ -923,7 +1000,7 @@ gst_queue_set_property (GObject * object,
 
   /* someone could change levels here, and since this
    * affects the get/put funcs, we need to lock for safety. */
-  g_mutex_lock (queue->qlock);
+  GST_QUEUE_MUTEX_LOCK;
 
   switch (prop_id) {
     case ARG_MAX_SIZE_BYTES:
@@ -958,7 +1035,7 @@ gst_queue_set_property (GObject * object,
       break;
   }
 
-  g_mutex_unlock (queue->qlock);
+  GST_QUEUE_MUTEX_UNLOCK;
 }
 
 static void
